"""Parent module of all the Nuclear physics related modules

Exports
-------
NuclearDataMap, NucleusData, generate_nucleus_id

Modules
-------
target
    Contains classes and methods for GasTarget and SolidTarget energy loss analysis
momentum
    Contains methods for momentum 4-vector analysis
"""

from .nuclear_map import NuclearDataMap, NucleusData, generate_nucleus_id



"""
The parent module of spyral-utils

Modules
-------
constants
    Useful constants for AT-TPC
plot
    Plotting utilities such as Cut2D and Histogrammer
nuclear
    Nuclear physics tools like nuclear masses, momentum 4-vectors, energy loss, etc.
"""



"""Module conatining methods for creating, saving, and using graphical Cuts on data

Classes
-------
CutHandler
    Handler to recieve vertices from a matplotlib selector (i.e. PolygonSelector).
Cut2D
    Implementation of 2D cuts as used in many types of graphical analyses

Functions
---------
write_cut_json(cut: Cut2D, filepath: Path) -> bool
    Write the JSON representation of a Cut2D to a file
load_cut_json(filepath: Path) -> Cut2D | None
    Deserialize the JSON representation of a Cut2D
"""
from matplotlib.path import Path as mplPath
from polars import Series
import numpy as np
import json
from pathlib import Path


class CutHandler:
    """Handler to recieve vertices from a matplotlib selector (i.e. PolygonSelector).

    Typically will be used interactively. The onselect method should be passed to the selector object at construction.
    CutHandler can also be used in analysis applications to store cuts.

    An example script:

    ```python
    from spyral_utils.plot import CutHandler, Cut2D, write_cut_json
    from matplotlib.widgets import PolygonSelector
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(1,1)
    handler = CutHandler()
    selector = PolygonSelector(ax, handler.onselect)

    #plot some data here...

    plt.show()

    #wait for user to draw a cut and close the window

    mycut = handler.cuts['cut_0']
    mycut.name = 'mycut'
    write_cut_json(mycut, 'mycut.json')
    ```

    Attributes
    ----------
    cuts: dict[str, Cut2D]
        mapping of cut name to Cut2D

    Methods
    -------
    onselect(verticies: list[tuple[float, float]])
        recieve a matplotlib polygon and create a Cut2D from it
    """

    def __init__(self):
        self.cuts: dict[str, Cut2D] = {}

    def onselect(self, vertices: list[tuple[float, float]]):
        cut_default_name = f"cut_{len(self.cuts)}"
        self.cuts[cut_default_name] = Cut2D(cut_default_name, vertices)


class Cut2D:
    """Implementation of 2D cuts as used in many types of graphical analyses

    Uses matplotlib Path objects. Takes in a name (to identify the cut) and a list of points. The Path
    takes the verticies, and can then be used to check if a point(s) is inside of the polygon using the
    is_*_inside functions. Can be serialized to json format. Can also retreive Nx2 ndarray of vertices
    for plotting after the fact.

    Attributes
    ----------
    path: matplotlib.path.Path
        A matplotlib path (polygon) that is the actual cut shape
    name: str
        A name for the cut

    Methods
    -------
    is_point_inside(x: float, y: float) -> bool
        Check if a single point (x,y) is inside the cut
    is_arr_inside(points: list[tuple[float, float]]) -> list[bool]
        Check if a list of points (x,y) are inside the cut
    is_cols_inside(columns: Series) -> Series
        Check if a set of polars Columns are inside the cut
    get_vertices() -> ndarray
        Get the cut vertices
    to_json_str() -> str
        Get the JSON representation of the cut
    """

    def __init__(self, name: str, vertices: list[tuple[float, float]]):
        self.path: mplPath = mplPath(
            vertices, closed=False
        )  # Has to be false, sometimes without this, the algorithm does some weird jumping between the first and last point
        self.name = name

    def is_point_inside(self, x: float, y: float) -> bool:
        """Is a point in the cut

        Parameters
        ----------
        x: float
            point x-coordinate
        y: float
            point y-coordinate

        Returns
        -------
        bool
            true if inside, false if outside
        """
        return self.path.contains_point((x, y))

    def is_arr_inside(self, points: list[tuple[float, float]]) -> list[bool]:
        """Which of the points in this list are in the cut

        Parameters
        ----------
        points: list[tuple[float, float]]
            List of points (x,y)

        Returns
        -------
        list[bool]
            List of results of checking each point
        """
        return self.path.contains_points(points)

    def is_cols_inside(self, columns: Series) -> Series:
        """Which of the points in this Series are in the cut

        Parameters
        ----------
        columns: Series
            Polars dataframe series to check

        Returns
        -------
        Series
            Series of True or False for each point
        """
        data = np.transpose(
            [columns.struct.field(name).to_list() for name in columns.struct.fields]
        )
        return Series(values=self.path.contains_points(data))

    def get_vertices(self) -> np.ndarray:
        """Get the cut vertices

        Returns
        -------
        ndarray
            the vertices

        """
        return self.path.vertices

    def to_json_str(self) -> str:
        """Get the cut JSON representation

        Returns
        -------
        str
            JSON representation
        """
        return json.dumps(
            self,
            default=lambda obj: {
                "name": obj.name,
                "vertices": obj.path.vertices.tolist(),
            },
            indent=4,
        )


def write_cut_json(cut: Cut2D, filepath: Path) -> bool:
    """Write the JSON representation of a Cut2D to a file

    Parameters
    ----------
    cut: Cut2D
        Cut to serialize
    filepath: Path
        Path at which cut should be written

    Returns
    -------
    bool
        True on success, False on failure
    """
    json_str = cut.to_json_str()
    try:
        with open(filepath, "w") as output:
            output.write(json_str)
            return True
    except OSError as error:
        print(f"An error occurred writing cut {cut.name} to file {filepath}: {error}")
        return False


def load_cut_json(filepath: Path) -> Cut2D | None:
    """Deserialize the JSON representation of a Cut2D

    Parameters
    ----------
    filepath: Path
        Path at which cut should be read from

    Returns
    -------
    Cut2D | None
        Returns a Cut2D on success, None on failure
    """
    try:
        with open(filepath, "r") as input:
            buffer = input.read()
            cut_dict = json.loads(buffer)
            if "name" not in cut_dict or "vertices" not in cut_dict:
                print(
                    f"Data in file {filepath} is not the right format for Cut2D, could not load"
                )
                return None
            return Cut2D(cut_dict["name"], cut_dict["vertices"])
    except OSError as error:
        print(
            f"An error occurred reading trying to read a cut from file {filepath}: {error}"
        )
        return None



"""
Module of some common constants aliased from scipy to avoid long key strings and with units most
relevant to AT-TPC data.

Attributes
---------
AMU_2_MEV
    converts u to MeV/c^2 (MeV/c^2 / u)
ELECTRON_MASS_U
    mass of an electron in amu (amu)
MEV_2_JOULE
    Convert MeV/c^2 to J (J/MeV/c^2)
MEV_2_KG
    Convert MeV/c^2 to kg (kg/MeV/c^2)
ROOM_TEMPERATURE
    Room temperature in K (K)
GAS_CONSTANT
    The gas constant (cm^3 Torr / (K mol))
QBRHO_2_P
    Constant used to convert magnetic rigidity times charge (unitless) to momentum (MeV/c / (T*m))
C
    speed of light (m/s)
"""

from scipy.constants import physical_constants, speed_of_light

AMU_2_MEV: float = physical_constants["atomic mass constant energy equivalent in MeV"][
    0
]  # CODATA 2018, convert u to MeV/c^2

ELECTRON_MASS_U: float = physical_constants["electron mass in u"][
    0
]  # CODATA 2018, evaluated by scipy

MEV_2_JOULE: float = (
    physical_constants["electron volt-joule relationship"][0] * 1.0e6
)  # J/ev * ev/MeV = J/MeV

MEV_2_KG: float = (
    physical_constants["electron volt-kilogram relationship"][0] * 1.0e6
)  # kg/ev * ev/MeV = kg/MeV (per c^2)

ROOM_TEMPERATURE: float = 293.0  # Kelvin

GAS_CONSTANT: float = (
    physical_constants["molar gas constant"][0] * 0.0075 * ((100.0) ** 3.0)
)  # m^3 Pa / K mol -> m^3 Torr / K mol -> cm^3 Torr / K mol

QBRHO_2_P: float = 1.0e-9 * 10.0 * 100.0 * speed_of_light  # T * m -> MeV/c

C: float = speed_of_light  # m/s



from spyral_utils.nuclear.target import load_target, GasTarget, SolidTarget
from spyral_utils.nuclear import NuclearDataMap
from pathlib import Path
import numpy as np

GAS_TARGET_PATH: Path = Path(__file__).parent.resolve() / "gas_target.json"
SOLID_TARGET_PATH: Path = Path(__file__).parent.resolve() / "solid_target.json"
PROJ_Z: int = 1
PROJ_A: int = 1
PROJ_KE: float = 5.0  # MeV


def test_gas_target():
    PRECISION = 0.01  # 5% of LISE value
    LISE_VALUE = 0.655  # MeV

    nuc_map = NuclearDataMap()
    target = load_target(GAS_TARGET_PATH, nuc_map)

    assert isinstance(target, GasTarget)
    assert target.ugly_string == "(Gas)1H2"

    proj_data = nuc_map.get_data(PROJ_Z, PROJ_A)

    eloss = target.get_energy_loss(proj_data, PROJ_KE, np.array([1.0]))

    assert abs(eloss[0] - LISE_VALUE) < PRECISION * LISE_VALUE


def test_solid_target():
    PRECISION = 0.01  # 5% of LISE value
    LISE_VALUE = 0.00512  # MeV

    nuc_map = NuclearDataMap()
    target = load_target(SOLID_TARGET_PATH, nuc_map)

    assert isinstance(target, SolidTarget)
    assert target.ugly_string == "(Solid)12C1"

    proj_data = nuc_map.get_data(PROJ_Z, PROJ_A)

    eloss = target.get_energy_loss(proj_data, PROJ_KE, np.array([np.pi * 0.25]))

    assert abs(eloss[0] - LISE_VALUE) < PRECISION * LISE_VALUE



"""Module containing AT-TPC representation of target materials and their associated energy loss calculations.

Classes should never be directly instantiated, rather use the load_target function to do the work for you.

Classes
-------
TargetData
    Raw target dataclass, should never be instantiated directly by user
GasTarget
    AT-TPC representation of a gas target
SolidTarget
    AT-TPC representation of a gas target

Functions
---------
load_target_data(target_path: Path, nuclear_map: NuclearDataMap) -> TargetData | None
    Load raw target data from JSON. Should never be used directly.
save_target_data(target_path: Path, target_data: TargetData)
    Save raw target data to JSON. Should never be used directly.
load_target(target_path: Path, nuclear_map: NuclearDataMap) -> GasTarget | SolidTarget | None
    Load a target from JSON. This is what should be used to load a target.
save_target(target_path: Path, target: GasTarget | SolidTarget)
    Save a targe to JSON. This is what should be used to save a target.
"""

from .nuclear_map import NuclearDataMap, NucleusData
from ..constants import AMU_2_MEV, GAS_CONSTANT, ROOM_TEMPERATURE

import pycatima as catima
from dataclasses import dataclass, field
from pathlib import Path
from json import load, dumps
import numpy as np


@dataclass
class TargetData:
    """Raw target dataclass

    Attributes
    ----------
    compound: list[tuple[int, int, int]]
        list of Z, A, S for each element in the compound
    pressure: float | None
        optional pressure of the gas in Torr
    thickness: float | None
        optional thickness of the solid in ug/cm^2

    Methods
    -------
    density() -> float
        get the density of the gas (return 0 if solid as this should not be used)
    """

    compound: list[tuple[int, int, int]] = field(default_factory=list)  # (Z, A, S)
    pressure: float | None = None  # torr
    thickness: float | None = None  # ug/cm^2

    def density(self) -> float:
        """Get the gas density in g/cm^3

        Returns
        -------
        float
            density of the gas in g/cm^3 (0 if the target is solid)
        """
        if self.pressure is None:
            return 0.0
        else:
            molar_mass: float = 0.0
            for z, a, s in self.compound:
                molar_mass += a * s
            return molar_mass * self.pressure / (GAS_CONSTANT * ROOM_TEMPERATURE)


def load_target_data(
    target_path: Path, nuclear_map: NuclearDataMap
) -> TargetData | None:
    """Load the raw target data from a JSON file

    Parameters
    ----------
    target_path: Path
        Path to JSON file containing target data
    nuclear_map: NuclearDataMap
        Mass data; unused

    Returns
    -------
    TargetData | None
        If the data loaded successfully, returns a TargetData. Otherwise returns None.
    """
    with open(target_path, "r") as target_file:
        json_data = load(target_file)
        if (
            "compound" not in json_data
            or "pressure(Torr)" not in json_data
            or "thickness(ug/cm^2)" not in json_data
        ):
            return None
        else:
            return TargetData(
                json_data["compound"],
                json_data["pressure(Torr)"],
                json_data["thickness(ug/cm^2)"],
            )


def save_target_data(target_path: Path, data: TargetData):
    """Write raw target data to JSON file

    Parameters
    ----------
    target_path: Path
        Path to JSON file to write target data
    data: TargetData
        data to be written
    """
    with open(target_path, "w") as target_file:
        json_str = dumps(
            data,
            default=lambda data: {
                "compound": data.compound,
                "pressure(Torr)": data.pressure,
                "thickness(ug/cm^2)": data.thickness,
            },
        )
        target_file.write(json_str)


class GasTarget:
    """An AT-TPC gas target

    Gas target for which energy loss can be calculated using pycatima. Can perform several types of
    energy loss calculations (straggling, dEdx, energy lost, etc.)

    Attributes
    ----------
    data: TargetData
        The raw target data from a JSON file
    ugly_string: str
        A string representation without rich formatting
    pretty_string: str
        A string representation with rich formatting
    material: catima.Material
        The catima representation of the target
    density: float
        The target density in g/cm^3

    Methods
    -------
    get_dedx(projectile_data: NucleusData, projectile_energy: float) -> float
        get the stopping power (dEdx) for a projectile in this target
    get_angular_straggling(projectile_data: NucleusData, projectile_energy: float) -> float:
        get the angular straggling for a projectile in this target
    get_energy_loss(projectile_data: NucleusData, projectile_energy: float, distances: ndarray) -> ndarray:
        get the energy loss values for a projectile travelling distances through the target
    """

    def __init__(self, target_data: TargetData, nuclear_data: NuclearDataMap):
        self.data = target_data

        self.pretty_string: str = "(Gas)" + "".join(
            [
                f"{nuclear_data.get_data(z, a).pretty_iso_symbol}<sub>{s}</sub>"
                for (z, a, s) in self.data.compound
            ]
        )
        self.ugly_string: str = "(Gas)" + "".join(
            [
                f"{nuclear_data.get_data(z, a).isotopic_symbol}{s}"
                for (z, a, s) in self.data.compound
            ]
        )

        # Construct the target material
        self.material = catima.Material()
        for (
            z,
            a,
            s,
        ) in self.data.compound:
            self.material.add_element(
                nuclear_data.get_data(z, a).atomic_mass, z, float(s)
            )
        self.density: float = self.data.density()
        self.material.density(self.density)

    def __str__(self) -> str:
        return self.pretty_string

    def get_dedx(self, projectile_data: NucleusData, projectile_energy: float) -> float:
        """Calculate the stopping power of the target for a projectile

        Parameters
        ----------
        projectile_data: NucleusData
            the projectile type
        projectile_energy: float
            the projectile kinetic energy in MeV

        Returns
        -------
        float
            dEdx in MeV/g/cm^2
        """
        mass_u = projectile_data.mass / AMU_2_MEV  # convert to u
        projectile = catima.Projectile(mass_u, projectile_data.Z)
        projectile.T(projectile_energy / mass_u)
        return catima.dedx(projectile, self.material)

    def get_angular_straggling(
        self, projectile_data: NucleusData, projectile_energy: float
    ) -> float:
        """Calculate the angular straggling for a projectile

        Parameters
        ----------
        projectile_data: NucleusData
            the projectile type
        projectile_energy: float
            the projectile kinetic energy in MeV

        Returns
        -------
        float
            angular straggling in radians
        """
        mass_u = projectile_data.mass / AMU_2_MEV  # convert to u
        projectile = catima.Projectile(
            mass_u, projectile_data.Z, T=projectile_energy / mass_u
        )
        return catima.calculate(projectile, self.material).get_dict()["sigma_a"]

    def get_energy_loss(
        self,
        projectile_data: NucleusData,
        projectile_energy: float,
        distances: np.ndarray,
    ) -> np.ndarray:
        """
        Calculate the energy loss of a projectile traveling over a set of distances

        Parameters
        ----------
        projectile_data: NucleusData
            the projectile type
        projectile_energy: float
            the projectile kinetic energy in MeV
        distances: ndarray
            a set of distances in meters over which to calculate the energy loss

        Returns
        -------
        ndarray
            set of energy losses
        """
        mass_u = projectile_data.mass / AMU_2_MEV  # convert to u
        projectile = catima.Projectile(
            mass_u, projectile_data.Z, T=projectile_energy / mass_u
        )
        eloss = np.zeros(len(distances))
        for idx, distance in enumerate(distances):
            self.material.thickness_cm(distance * 100.0)
            projectile.T(projectile_energy / mass_u)
            eloss[idx] = catima.calculate(projectile, self.material).get_dict()["Eloss"]
        return eloss


class SolidTarget:
    """An AT-TPC gas target

    Gas target for which energy loss can be calculated using pycatima. Can perform several types of
    energy loss calculations (straggling, dEdx, energy lost, etc.)

    Attributes
    ----------
    data: TargetData
        The raw target data from a JSON file
    ugly_string: str
        A string representation without rich formatting
    pretty_string: str
        A string representation with rich formatting
    material: catima.Material
        The catima representation of the target

    Methods
    -------
    get_dedx(projectile_data: NucleusData, projectile_energy: float) -> float
        get the stopping power (dEdx) for a projectile in this target
    get_angular_straggling(projectile_data: NucleusData, projectile_energy: float) -> float:
        get the angular straggling for a projectile in this target
    get_energy_loss(projectile_data: NucleusData, projectile_energy: float, distances: ndarray) -> ndarray:
        get the energy loss values for a projectile travelling distances through the target
    """

    UG2G: float = 1.0e-6  # convert ug to g

    def __init__(self, target_data: TargetData, nuclear_data: NuclearDataMap):
        self.data = target_data
        self.pretty_string: str = "(Solid)" + "".join(
            [
                f"{nuclear_data.get_data(z, a).pretty_iso_symbol}<sub>{s}</sub>"
                for (z, a, s) in self.data.compound
            ]
        )
        self.ugly_string: str = "(Solid)" + "".join(
            [
                f"{nuclear_data.get_data(z, a).isotopic_symbol}{s}"
                for (z, a, s) in self.data.compound
            ]
        )

        self.material = catima.Material()
        for (
            z,
            a,
            s,
        ) in self.data.compound:
            self.material.add_element(
                nuclear_data.get_data(z, a).atomic_mass, z, float(s)
            )
        self.material.thickness(
            target_data.thickness * self.UG2G
        )  # Convert ug/cm^2 to g/cm^2

    def get_dedx(self, projectile_data: NucleusData, projectile_energy: float) -> float:
        """Calculate the stopping power of the target for a projectile

        Parameters
        ----------
        projectile_data: NucleusData
            the projectile type
        projectile_energy: float
            the projectile kinetic energy in MeV

        Returns
        -------
        float
            dEdx in MeV/g/cm^2
        """
        mass_u = projectile_data.mass / AMU_2_MEV  # convert to u
        projectile = catima.Projectile(mass_u, projectile_data.Z)
        projectile.T(projectile_energy / mass_u)
        return catima.dedx(projectile, self.material)

    def get_angular_straggling(
        self, projectile_data: NucleusData, projectile_energy: float
    ) -> float:
        """Calculate the angular straggling for a projectile

        Parameters
        ----------
        projectile_data: NucleusData
            the projectile type
        projectile_energy: float
            the projectile kinetic energy in MeV

        Returns
        -------
        float
            angular straggling in radians
        """
        mass_u = projectile_data.mass / AMU_2_MEV  # convert to u
        projectile = catima.Projectile(
            mass_u, projectile_data.Z, T=projectile_energy / mass_u
        )
        return catima.calculate(projectile, self.material).get_dict()["sigma_a"]

    def get_energy_loss(
        self,
        projectile_data: NucleusData,
        projectile_energy: float,
        incident_angles: np.ndarray,
    ) -> np.ndarray:
        """Calculate the energy loss of a projectile traveling through the solid target for a given set of incident angles

        Assumes travelling through the entire target at an incident angle.

        Parameters
        ----------
        projectile_data: NucleusData
            the projectile type
        projectile_energy: float
            the projectile kinetic energy in MeV
        incident_angles: ndarray
            a set of incident angles, describing the angle between the particle trajectory and the normal of the target surface

        Returns
        -------
        ndarray
            set of energy losses
        """
        mass_u = projectile_data.mass / AMU_2_MEV  # convert to u
        projectile = catima.Projectile(
            mass_u, projectile_data.Z, T=projectile_energy / mass_u
        )
        eloss = np.zeros(len(incident_angles))
        nominal_thickness = self.material.thickness()
        for idx, angle in enumerate(incident_angles):
            self.material.thickness(nominal_thickness / abs(np.cos(angle)))
            projectile.T(projectile_energy / mass_u)
            eloss[idx] = catima.calculate(projectile, self.material).get_dict()["Eloss"]
        self.material.thickness(nominal_thickness)
        return eloss


def load_target(
    target_path: Path, nuclear_map: NuclearDataMap
) -> GasTarget | SolidTarget | None:
    """Load a target from a JSON file

    Read the JSON data, and construct the appropriate target type.

    Parameters
    ----------
    target_path: Path
        Path to JSON data
    nuclear_map: NuclearDataMap
        Nucleus mass data

    Returns
    -------
    GasTarget | SolidTarget | None
        Return a GasTarget or SolidTarget where appropriate. Return None on failure.
    """
    data = load_target_data(target_path, nuclear_map)
    if data is None:
        return None
    elif data.pressure is None:
        return SolidTarget(data, nuclear_map)
    else:
        return GasTarget(data, nuclear_map)


def save_target(target_path: Path, target: GasTarget | SolidTarget):
    """Write a target to a JSON file

    Create the JSON data, and write to disk.

    Parameters
    ----------
    target_path: Path
        Path to JSON data file
    target: GasTarget | SolidTarget
        Target to be written
    """
    save_target_data(target_path, target.data)



"""Modules for creating and plotting histograms in matplotlib

The default matplotlib histogramming tools can be hard to use for large datasets
as they require a lot of rebinning or all of the data to be loaded. Here we give some of our own
which allow for incremental filling of histograms

Classes
-------
Hist1D
    A 1-D histogram dataclass. Should not be instantiated directly
Hist2D
    A 2-D histogram dataclass. Should not be instantiated directly
Histogrammer
    A parent object used to create, manage, draw histograms
"""

import numpy as np
from numpy.typing import NDArray
from dataclasses import dataclass
from matplotlib.axes import Axes
from matplotlib.text import Text
from matplotlib.colors import LogNorm
from matplotlib.backend_bases import LocationEvent
from matplotlib.collections import QuadMesh
import matplotlib.pyplot as plt
from matplotlib import colormaps
from math import floor

CMAP = colormaps.get_cmap("viridis").with_extremes(under="white")


# Utility functions
def clamp_low(x: float, edge: float) -> float:
    return x if x > edge else edge


def clamp_hi(x: float, edge: float) -> float:
    return x if x < edge else edge


def clamp_range(xrange: tuple[float, float], min_max: tuple[float, float]):
    return (clamp_low(xrange[0], min_max[0]), clamp_hi(xrange[1], min_max[1]))


@dataclass
class Hist1D:
    """Dataclass wrapping a numpy array used to store histogram data and retrieve histogram statistics

    Attributes
    ----------
    name: str
        histogram name
    counts: ndarray
        array of histogram counts
    bins: ndarray
        array of histogram bin edges
    bin_width: float
        the width of histogram bins

    Methods
    -------
    get_bin(x: float) -> int | None
        get the bin number for an x-coordinate value
    stats_for_range(xrange: tuple[float, float]) -> tuple[float, float, float] | None
        get some statistics for a subrange of the histogram
    get_subrange(xrange: tuple[float, float]) -> tuple[ndarray, ndarray]
        get a histogram subrange (bin edges, counts)
    """

    name: str
    counts: NDArray[np.float64]
    bins: NDArray[np.float64]
    bin_width: float

    def get_bin(self, x: float) -> int | None:
        """Get the bin number which contains the x-coordinate

        Parameters
        ----------
        x: float
            X-coordinate for which we want to find the bin number

        Returns
        -------
        int | None
            The bin number or None if the x value does not fall within the histogram
        """
        if x < self.bins.min() or x > self.bins.max():
            return None

        return int(floor((x - self.bins[0]) / self.bin_width))

    def stats_for_range(
        self, xrange: tuple[float, float]
    ) -> tuple[float, float, float] | None:
        """Get some statistics for a histogram subrange

        Calculates the mean, integral, and standard deviation of the sub-range

        Parameters
        ----------
        xrange: tuple[float, float]
            the subrange of the histogram (min, max) in x-coordinates

        Returns
        -------
        tuple[float, float, float] | None
            Returns a tuple of (integral, mean, std. dev.) for the subrange, or None if the subrange is not within the histogram bounds

        """
        clamped_range = clamp_range(xrange, (self.bins.min(), self.bins.max()))
        bin_min = self.get_bin(clamped_range[0])
        bin_max = self.get_bin(clamped_range[1])
        if bin_min is None or bin_max is None:
            return None
        integral = np.sum(self.counts[bin_min:bin_max])
        mean = np.average(
            self.bins[bin_min:bin_max], weights=self.counts[bin_min:bin_max]
        )
        variance = np.average(
            (self.bins[bin_min:bin_max] - mean) ** 2.0,
            weights=self.counts[bin_min:bin_max],
        )
        return (integral, mean, np.sqrt(variance))

    def get_subrange(
        self, xrange: tuple[float, float]
    ) -> tuple[np.ndarray, np.ndarray]:
        """Get a subrange of the histogram

        Parameters
        ----------
        xrange: tuple[float, float]
            the subrange of the histogram (min, max) in x-coordinates

        Returns
        -------
        tuple[ndarray, ndarray]
            the subrange (bin edges, counts)
        """
        mask = np.logical_and(self.bins > xrange[0], self.bins < xrange[1])
        return (self.bins[mask], self.counts[mask[:-1]])


@dataclass
class Hist2D:
    """Dataclass wrapping a numpy array used to store two-dimensional histogram data and retrieve histogram statistics

    Attributes
    ----------
    name: str
        histogram name
    counts: ndarray
        array of histogram counts
    x_bins: ndarray
        array of histogram x bin edges
    y_bins: ndarray
        array of histogram y bin edges
    x_bin_width: float
        the width of histogram x bins
    y_bin_width: float
        the width of histogram y bins

    Methods
    -------
    get_bin(coords: tuple[float, float]) -> tuple[int, int] | None
        get the x and y bin numbers for an (x,y)-coordinate value
    stats_for_range(xrange: tuple[float, float], yrange: tuple[float, float]) -> tuple[float, float, float, float, float] | None
        get some statistics for a subrange of the histogram
    get_subrange(xrange: tuple[float, float], xrange: tuple[float, float]) -> tuple[ndarray, ndarray, ndarray]
        get a histogram subrange (x bin edges, y bin edges, counts)
    """

    name: str
    counts: NDArray[np.float64]
    x_bins: NDArray[np.float64]
    y_bins: NDArray[np.float64]
    x_bin_width: float
    y_bin_width: float

    def get_bin(self, coords: tuple[float, float]) -> tuple[int, int] | None:
        """Get the x and y bin numbers for an (x,y)-coordinate value

        Parameters
        ----------
        coords: tuple[float, float]
            The (x,y) corrdinate for which we want to find the bin numbers

        Returns
        -------
        tuple[int, int] | None
            Returns the (x bin, y bin) numbers or None if out of range
        """
        if (coords[0] < self.x_bins.min() or coords[0] > self.x_bins.max()) or (
            coords[1] < self.y_bins.min() or coords[1] > self.y_bins.max()
        ):
            return None

        y_bin = int(floor((coords[1] - self.y_bins[0]) / self.y_bin_width))
        x_bin = int(floor((coords[0] - self.x_bins[0]) / self.x_bin_width))
        return (x_bin, y_bin)

    # returns (integral, mean x, std_dev x, mean y, std_dev y)
    def stats_for_range(
        self, xrange: tuple[float, float], yrange: tuple[float, float]
    ) -> tuple[float, float, float, float, float] | None:
        """Get some statistics for a histogram subrange

        Calculates the mean in x and y, integral, and standard deviation in x and y of the sub-range

        Parameters
        ----------
        xrange: tuple[float, float]
            the subrange of the histogram (min, max) in x-coordinates
        yrange: tuple[float, float]
            the subrange of the histogram (min, max) in y-coordinates

        Returns
        -------
        tuple[float, float, float, float, float] | None
            Returns a tuple of (integral, x mean, y mean, x std. dev., y std. dev.) for the subrange, or None if the subrange is not within the histogram bounds

        """
        clamped_x_range = clamp_range(xrange, (self.x_bins.min(), self.x_bins.max()))
        clamped_y_range = clamp_range(yrange, (self.y_bins.min(), self.y_bins.max()))
        bin_min = self.get_bin((clamped_x_range[0], clamped_y_range[0]))
        bin_max = self.get_bin((clamped_x_range[1], clamped_y_range[1]))

        x_bin_range = np.arange(start=bin_min[0], stop=bin_max[0], step=1)
        y_bin_range = np.arange(start=bin_min[1], stop=bin_max[1], step=1)
        bin_mesh = np.ix_(y_bin_range, x_bin_range)

        integral = np.sum(self.counts[bin_mesh])
        mean_x = np.average(
            self.x_bins[bin_min[0] : bin_max[0]],
            weights=np.sum(self.counts.T[bin_min[0] : bin_max[0]], 1),
        )
        mean_y = np.average(
            self.y_bins[bin_min[1] : bin_max[1]],
            weights=np.sum(self.counts[bin_min[1] : bin_max[1]], 1),
        )
        var_x = np.average(
            (self.x_bins[bin_min[0] : bin_max[0]] - mean_x) ** 2.0,
            weights=np.sum(self.counts.T[bin_min[0] : bin_max[0]], 1),
        )
        var_y = np.average(
            (self.y_bins[bin_min[1] : bin_max[1]] - mean_y) ** 2.0,
            weights=np.sum(self.counts[bin_min[1] : bin_max[1]], 1),
        )
        return (integral, mean_x, mean_y, np.sqrt(var_x), np.sqrt(var_y))

    def get_subrange(
        self, xrange: tuple[float, float], yrange: tuple[float, float]
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Get a subrange of the histogram

        Parameters
        ----------
        xrange: tuple[float, float]
            the subrange of the histogram (min, max) in x-coordinates
        yrange: tuple[float, float]
            the subrange of the histogram (min, max) in y-coordinates

        Returns
        -------
        tuple[ndarray, ndarray, ndarray]
            the subrange (x bin edges, y bin edges, counts)
        """
        x_mask = np.logical_and(self.x_bins > xrange[0], self.x_bins < xrange[1])
        y_mask = np.logical_and(self.y_bins > yrange[0], self.y_bins < yrange[1])
        bin_mesh = np.ix_(y_mask, x_mask)
        return (self.x_bins[x_mask], self.y_bins[y_mask], self.counts[bin_mesh])


class Histogrammer:
    """Histogrammer is a wrapper around a dictionary of str->Hist1D|Hist2D that interfaces with matplotlib

    A new histogram can be added to the dictionary using the add_hist1d/add_hist2d methods. The name passed to
    these methods is used as the key for the dictionary. To add data to the histograms use the fill_hist1d/fill_hist2d methods.
    The fill methods accept arrays of data, and this is by intent. It would not be efficient to fill the histograms point by point. Rather, prefer
    passing entire data sets (like dataframe columns). Finally, to retrieve a histogram (for plotting, etc), use the get_hist1d/get_hist2d methods.
    Prefer the getters over direct access to the underlying dictionary as the getters perfom some error checking.

    Attributes
    ----------
    histograms: dict[str, Hist1D | Hist2D]
        the histograms held by the Histogrammer, mapped by name
    axes: dict[Axes, tuple[str, Text | None]]
        mapping of mpl Axes to associated histogram names and Text for the stats
    figures: dict[str, bool]
        used to see if callbacks have been set for a given figure name

    Methods
    -------
    add_hist1d(name: str, bins: int, range: tuple[float, float])
        add a Hist1D
    add_hist2d(name: str, bins: tuple[int, int], ranges: tuple[tuple[float, float], tuple[float, float]])
        add a Hist2D
    fill_hist1d(self, name: str, data: ndarray) -> bool
        fill an existing Hist1D with some data
    fill_hist2d(self, name: str, x_data: ndarray, y_data: ndarray) -> bool
        fill an existing Hist2D with some data
    get_hist1d(name: str) -> Hist1D | None
        get a Hist1D by name
    get_hist2d(name: str) -> Hist2D | None
        get a Hist2D by name
    on_axes_enter_hist1d(event: LocationEvent)
        handler for when the user mouse enters an axis of a Hist1D
    on_axes_enter_hist2d(event: LocationEvent)
        handler for when the user mouse enters an axis of a Hist2D
    on_axes_enter(event: LocationEvent)
        matplotlib callback for when the user mouse enters an axis
    on_axes_leave(event: LocationEvent)
        matplotlib callback for when the user mouse leaves an axis
    connect_mpl_callbacks(axis: Axes)
        setup mpl callbacks for an axis
    draw_hist1d(name: str, axis: Axes)
        draw a Hist1D in a matplotlib Axes
    draw_hist2d(name: str, axis: Axes, log_z: bool = False) -> QuadMesh | None
        draw a Hist2D in a matplotlib Axes
    """

    def __init__(self):
        self.histograms: dict[str, Hist1D | Hist2D] = {}
        self.axes: dict[Axes, tuple[str, Text | None]] = {}
        self.figures: dict[
            str, bool
        ] = {}  # used to indicate if callbacks have been bound for that figure

    def add_hist1d(self, name: str, bins: int, range: tuple[float, float]):
        """Add a Hist1D to the Histogrammer

        Parameters
        ----------
        name: str
            The name of the histogram, it should be unqiue
        bins: int
            The number of bins
        range: tuple[float, float]
            The x-range of the histogram in x-axis coordinates
        """
        if name in self.histograms:
            print(f"Overwriting histogram named {name} in Histogrammer.add_histogram!")

        hist = Hist1D(
            name, np.empty(0), np.empty(0), np.abs(range[0] - range[1]) / float(bins)
        )
        hist.counts, hist.bins = np.histogram(a=[], bins=bins, range=range)
        self.histograms[name] = hist

    def add_hist2d(
        self,
        name: str,
        bins: tuple[int, int],
        ranges: tuple[tuple[float, float], tuple[float, float]],
    ):
        """Add a Hist2D to the Histogrammer

        Parameters
        ----------
        name: str
            The name of the histogram, it should be unqiue
        bins: tuple[int, int]
            The number of (x bins, y bins)
        ranges: tuple[tuple[float, float], tuple[float, float]]
            The range of the histogram ((min x, max x), (min y, max y))
        """
        if name in self.histograms:
            print(f"Overwriting histogram named {name} in Histogrammer.add_histogram!")

        hist = Hist2D(
            name,
            np.empty(0),
            np.empty(0),
            np.empty(0),
            np.abs(ranges[0][0] - ranges[0][1]) / float(bins[0]),
            np.abs(ranges[1][0] - ranges[1][1]) / float(bins[1]),
        )
        hist.counts, hist.x_bins, hist.y_bins = np.histogram2d(
            x=[], y=[], bins=bins, range=ranges
        )
        hist.counts = hist.counts.T
        self.histograms[name] = hist

    def fill_hist1d(self, name: str, data: np.ndarray) -> bool:
        """Fill a Hist1D with some data

        Parameters
        ----------
        name: str
            The name of the Hist1D
        data: ndarray
            The data to fill the histogram with. Should be a 1-D array

        Returns
        -------
        bool
            Indicates if data was successfully added to the histogram

        """
        if name not in self.histograms:
            return False

        hist = self.histograms[name]
        if type(hist) is not Hist1D:
            return False

        hist.counts = hist.counts + np.histogram(a=data, bins=hist.bins)[0]
        return True

    def fill_hist2d(self, name: str, x_data: np.ndarray, y_data: np.ndarray) -> bool:
        """Fill a Hist1D with some data

        The parameters x_data and y_data should have the same length.

        Parameters
        ----------
        name: str
            The name of the Hist1D
        x_data: ndarray
            The x coordinates of the data. Should be a 1-D array.
        y_data: ndarray
            The y coordinates of the data. Should be a 1-D array.

        Returns
        -------
        bool
            Indicates if data was successfully added to the histogram

        """
        if name not in self.histograms:
            return False

        hist = self.histograms[name]
        if type(hist) is not Hist2D:
            return False
        counts, _, _ = np.histogram2d(
            x_data.flatten(), y_data.flatten(), bins=(hist.x_bins, hist.y_bins)
        )
        hist.counts += counts.T
        return True

    def get_hist1d(self, name: str) -> Hist1D | None:
        """Retrieve a Hist1D by name

        Parameters
        ----------
        name: str
            The name of the histogram

        Returns
        -------
        Hist1D | None
            Returns Hist1D if a Hist1D exists with the given name. Otherwise returns None.

        """
        if name not in self.histograms:
            return None

        hist = self.histograms[name]
        if type(hist) is not Hist1D:
            return None
        else:
            return hist

    def get_hist2d(self, name: str) -> Hist2D | None:
        """Retrieve a Hist2D by name

        Parameters
        ----------
        name: str
            The name of the histogram

        Returns
        -------
        Hist2D | None
            Returns Hist2D if a Hist2D exists with the given name. Otherwise returns None.

        """
        if name not in self.histograms:
            return None

        hist = self.histograms[name]
        if type(hist) is not Hist2D:
            return None
        else:
            return hist

    def on_axes_enter_hist1d(self, event: LocationEvent):
        """Handler for AxesEnter events on Hist1D

        Parameters
        ----------
        event: LocationEvent
            The event which triggered the callback

        """
        data = self.axes[event.inaxes]
        xrange = event.inaxes.get_xbound()
        yrange = event.inaxes.get_ybound()
        stats = self.histograms[data[0]].stats_for_range(xrange)
        if data[1] is not None:
            data[1].remove()

        draw_x = xrange[1] - 0.25 * np.abs(xrange[0] - xrange[1])
        draw_y = yrange[1] - 0.25 * np.abs(yrange[0] - yrange[1])
        self.axes[event.inaxes] = (
            data[0],
            event.inaxes.text(
                draw_x,
                draw_y,
                f"Integral: {stats[0]}\nCentroid: {stats[1]:.3f}\nSigma: {stats[2]:.3f}",
            ),
        )
        plt.draw()

    def on_axes_enter_hist2d(self, event: LocationEvent):
        """Handler for AxesEnter events on Hist2D

        Parameters
        ----------
        event: LocationEvent
            The event which triggered the callback

        """
        data = self.axes[event.inaxes]
        xrange = event.inaxes.get_xbound()
        yrange = event.inaxes.get_ybound()
        stats = self.histograms[data[0]].stats_for_range(xrange, yrange)
        if data[1] is not None:
            data[1].remove()

        draw_x = xrange[1] - 0.25 * np.abs(xrange[0] - xrange[1])
        draw_y = yrange[1] - 0.25 * np.abs(yrange[0] - yrange[1])
        self.axes[event.inaxes] = (
            data[0],
            event.inaxes.text(
                draw_x,
                draw_y,
                f"Integral: {stats[0]}\nCentroid X: {stats[1]:.3f}\nCentroid Y: {stats[2]:.3f}\nSigma X: {stats[3]:.3f}\nSigma Y: {stats[4]:.3f}",
                color="black",
            ),
        )
        plt.draw()

    def on_axes_enter(self, event: LocationEvent):
        """Callback function for AxesEnter events from matplotlib on histograms to display stats

        Parameters
        ----------
        event: LocationEvent
            The event which triggered the callback

        """
        if event.inaxes not in self.axes:
            return

        if type(self.histograms[self.axes[event.inaxes][0]]) is Hist1D:
            self.on_axes_enter_hist1d(event)
        elif type(self.histograms[self.axes[event.inaxes][0]]) is Hist2D:
            self.on_axes_enter_hist2d(event)

    def on_axes_leave(self, event: LocationEvent):
        """Callback function for AxesLeave events from matplotlib on histograms to un-display stats

        Parameters
        ----------
        event: LocationEvent
            The event which triggered the callback

        """
        if event.inaxes not in self.axes:
            return
        data = self.axes[event.inaxes]
        if data[1] is None:
            return
        data[1].remove()
        self.axes[event.inaxes] = (data[0], None)
        plt.draw()

    def connect_mpl_callbacks(self, axis: Axes):
        """Setup callbacks for matplotlib with a given axis

        Parameters
        ----------
        axis: Axes
            The Axes to bind the callbacks to

        """
        if not hasattr(axis.figure, "_suptitle"):
            axis.figure.suptitle(f"Figure {len(self.figures)}")
        elif axis.figure._suptitle in self.figures:
            return

        self.figures[axis.figure._suptitle] = True
        axis.figure.canvas.mpl_connect("axes_enter_event", self.on_axes_enter)
        axis.figure.canvas.mpl_connect("axes_leave_event", self.on_axes_leave)

    def draw_hist1d(self, name: str, axis: Axes):
        """Draw a Hist1D on a given Axes

        Parameters
        ----------
        name: str
            The name of the histogram to draw
        axis: Axes
            The Axes into which the histogram should be drawn

        """
        if name not in self.histograms:
            return

        hist = self.histograms[name]
        if type(hist) is not Hist1D:
            return

        axis.stairs(hist.counts, hist.bins)
        self.axes[axis] = (name, None)
        self.connect_mpl_callbacks(axis)

    def draw_hist2d(
        self, name: str, axis: Axes, log_z: bool = False
    ) -> QuadMesh | None:
        """Draw a Hist2D on a given Axes

        Parameters
        ----------
        name: str
            The name of the histogram to draw
        axis: Axes
            The Axes into which the histogram should be drawn
        log_z: bool
            Whether or not to logscale the z-axis (color). Default is False.

        Returns
        -------
        QuadMesh | None
            If successfully drawn, returns a QuadMesh of the color scale used to later draw a color bar. Otherwise returns None.

        """
        if name not in self.histograms:
            return None

        hist = self.histograms[name]
        if type(hist) is not Hist2D:
            return None
        mesh = None
        if log_z:
            mesh = axis.pcolormesh(
                hist.x_bins, hist.y_bins, hist.counts, cmap=CMAP, norm=LogNorm()
            )
        else:
            mesh = axis.pcolormesh(
                hist.x_bins, hist.y_bins, hist.counts, cmap=CMAP, vmin=1.0e-6
            )
        self.axes[axis] = (name, None)
        self.connect_mpl_callbacks(axis)
        return mesh



"""Module for loading and accessing nuclear mass data

Reads in the bundled AMDC AME mass data (amdc_2020.txt) and loads into a dictionary.

Classes
-------
NucleusData
    dataclass representing AME mass data
NuclearDataMap
    interface for accessing mass data

Functions
---------
generate_nucleus_id(z: int, a: int) -> int:
    get a unqiue nucleus id
"""

from dataclasses import dataclass
from ..constants import AMU_2_MEV, ELECTRON_MASS_U
from pathlib import Path

DATA_PATH: Path = Path(__file__).parent.resolve() / "amdc_2020.txt"


@dataclass
class NucleusData:
    """Nucleus data from AME dataclass

    Contains information on nuclear masses and symbols

    Attributes
    ----------
    mass: float
        mass of the nucleus in MeV
    atomic_mass: float
        atomic mass in amu
    element_symbol: str
        atomic symbol (H, He, Li, etc.)
    isotopic_symbol: str
        isotopic symbol without formatting (1H, 2H, 3H, 4He, etc.)
    pretty_iso_symbol: str
        isotopic symbol with rich text formatting (<sup>1</sup>H, etc.)
    Z: int
        atomic number
    A: int
        mass number

    Methods
    -------
    __str__()
        string representation, isotopic_symbol
    get_latex_rep()
        get a LaTeX style represenation, for use with matplotlib etc.

    """

    mass: float = 0.0  # nuclear mass, MeV
    atomic_mass: float = 0.0  # atomic mass (includes electrons), amu
    element_symbol: str = ""  # Element symbol (H, He, Li, etc.)
    isotopic_symbol: str = ""  # Isotopic symbol w/o formating (1H, 2H, 3H, 4He, etc.)
    pretty_iso_symbol: str = (
        ""  # Isotopic symbol w/ rich text formating (<sup>1</sup>H, etc.)
    )
    Z: int = 0
    A: int = 0

    def __str__(self) -> str:
        return self.isotopic_symbol

    def get_latex_rep(self) -> str:
        """Get the LaTeX representation of the isotopic symbol

        Returns
        -------
        str
            a string of the isotopic symbol in LaTeX format
        """

        return "$^{" + str(self.A) + "}$" + self.element_symbol


# Szudzik pairing function, requires use of unsigned integers
def generate_nucleus_id(z: int, a: int) -> int:
    """get a unqiue nucleus id

    Use the Szudzik pairing function to generate a unique nucleus ID

    Parameters
    ----------
    z: int
        Nucleus atomic number
    a: int
        Nucleus mass number

    Returns
    -------
    int
        Unique identifier for this nucleus

    """
    return z * z + z + a if z == max(z, a) else a * a + z


class NuclearDataMap:
    """Maps nuclear numbers (Z,A) to mass information

    Reads in AME 2020 mass evaluation and creates a map of nucleus numbers (Z, A)
    to nuclear mass information

    Attributes
    ----------
    map: dict[int, NucleusData]
        dictionary mapping nucleus id's to associated data

    Methods
    -------
    get_data(z: int, a: int) -> NucleusData
        retrieve the mass data for a given nucleus
    """

    def __init__(self):
        self.map = {}

        with open(DATA_PATH) as data_file:
            data_file.readline()  # Header
            for line in data_file:
                entries = line.split()
                data = NucleusData()
                data.Z = int(entries[0])  # Column 1: Z
                data.A = int(entries[1])  # Column 2: A
                data.element_symbol = entries[2]  # Column 3: Element
                data.atomic_mass = float(entries[3])
                data.mass = (
                    float(entries[3]) - float(data.Z) * ELECTRON_MASS_U
                ) * AMU_2_MEV  # Remove electron masses to obtain nuclear masses, Column 4
                data.isotopic_symbol = f"{data.A}{entries[2]}"
                data.pretty_iso_symbol = f"<sup>{data.A}</sup>{entries[2]}"
                self.map[generate_nucleus_id(data.Z, data.A)] = data

    def get_data(self, z: int, a: int) -> NucleusData:
        """retrieve the mass data for a given nucleus

        Parameters
        ----------
        z: int
            atomic number
        a: int
            mass number

        Returns
        -------
        NucleusData
            associated mass data
        """
        return self.map[generate_nucleus_id(z, a)]



from spyral_utils.nuclear.momentum import (
    momentum_4_vec_from_cartesian,
    momentum_4_vec_from_spherical,
    get_transform_to_CoM,
    apply_transform,
)
from math import sin, cos
import numpy as np


def test_momentum_generation():
    p = 3.0
    polar = 0.45
    azim = 0.0
    energy = 9.0 + 1.0

    px = p * sin(polar) * cos(azim)
    py = p * sin(polar) * sin(azim)
    pz = p * cos(polar)

    cart_vec = momentum_4_vec_from_cartesian(px, py, pz, energy)
    pol_vec = momentum_4_vec_from_spherical(polar, azim, p, energy)

    assert np.array_equal(cart_vec, pol_vec)


def test_lorentz_transform():
    PRECISION = 1.0e-10

    p = 3.0
    polar = 0.45
    azim = 0.0
    energy = 9.0 + 1.0

    vec = momentum_4_vec_from_spherical(polar, azim, p, energy)

    transform = get_transform_to_CoM(vec)
    boosted = apply_transform(transform, vec)

    print(boosted)

    assert abs(boosted[0]) < PRECISION
    assert abs(boosted[1]) < PRECISION
    assert abs(boosted[2]) < PRECISION



"""Parent module of all of the plotting related modules

Exports
-------
Cut2D, CutHandler, load_cut_json, write_cut_json
Hist1D, Hist2D, Histogrammer
"""

from .cut import Cut2D, CutHandler, load_cut_json, write_cut_json
from .histogram import Hist1D, Hist2D, Histogrammer



from spyral_utils.plot import Cut2D, load_cut_json
from pathlib import Path

CUT_JSON_PATH: Path = Path(__file__).parent.resolve() / "cut.json"


def test_cut():
    cut = load_cut_json(CUT_JSON_PATH)

    assert isinstance(cut, Cut2D)
    assert cut.is_point_inside(0.5, 0.5)



"""Module conaining functions for using momentum 4-vectors

Create numpy arrays for doing 4-vector math

Functions
---------
momentum_4_vec_from_cartesian(px: float, py: float, pz: float, E: float) -> ndarray
    Create a momentum 4-vector from cartesian momenta, energy
momentum_4_vec_from_spherical(polar: float, azimuthal: float, p: float, E: float) -> ndarray
    Create a momentum 4-vector from spherical angles, momentum, energy
get_transform_to_CoM(momentum: ndarray) -> ndarray:
    Get the center of mass transformation for a 4-vector
apply_transform(transform: ndarray, momentum: ndarray) -> ndarray:
    Apply a Lorentz transform to a 4-vector
"""

import numpy as np


def momentum_4_vec_from_cartesian(
    px: float, py: float, pz: float, E: float
) -> np.ndarray:
    """Create a momentum 4-vector from cartesian momenta, energy

    Create a momentum 4-vector from momentum x,y,z and energy

    Parameters
    ----------
    px: float
        Momentum along the x-coordinate
    py: float
        Momentum along the y-coordinate
    pz: float
        Momentum along the z-coordinate
    E: float
        Total energy

    Returns
    -------
    ndarray
        an array representing the four-vector (length 4, dtype float)
    """
    return np.array([px, py, pz, E])


def momentum_4_vec_from_spherical(
    polar: float, azimuthal: float, p: float, E: float
) -> np.ndarray:
    """Create a momentum 4-vector from spherical angles, momentum, energy

    Create a momentum 4-vector from the momentum magnitude, energy, and spherical coordinates angles

    Parameters
    ----------
    polar: float
        polar angle in radians
    azimuthal: float
        azimuthal angle in radians
    p: float
        magnitude of the 3-momentum
    E: float
        Total energy

    Returns
    -------
    ndarray
        an array representing the four-vector (length 4, dtype float)
    """
    return np.array(
        [
            p * np.sin(polar) * np.cos(azimuthal),
            p * np.sin(polar) * np.sin(azimuthal),
            p * np.cos(polar),
            E,
        ]
    )


def get_transform_to_CoM(momentum: np.ndarray) -> np.ndarray:
    """Get the center of mass transformation for a 4-vector

    Parameters
    ----------
    momentum: np.ndarray
        a length 4 momentum vector

    Returns
    -------
    ndarray
        a 4x4 matrix representing the transformation to the center of mass frame

    """
    beta_vec = -1.0 * momentum[:3] / momentum[3]
    beta = np.linalg.norm(beta_vec)
    gamma = 1.0 / np.sqrt(1.0 - beta**2.0)
    bgamma = gamma * gamma / (1.0 + gamma)
    bxy = beta_vec[0] * beta_vec[1]
    bxz = beta_vec[0] * beta_vec[2]
    byz = beta_vec[1] * beta_vec[2]
    return np.array(
        [
            [
                1.0 + bgamma * beta_vec[0] ** 2.0,
                bgamma * bxy,
                bgamma * bxz,
                gamma * beta_vec[0],
            ],
            [
                bgamma * bxy,
                1.0 + bgamma * beta_vec[1] ** 2.0,
                bgamma * byz,
                gamma * beta_vec[1],
            ],
            [
                bgamma * bxz,
                bgamma * byz,
                1.0 + bgamma * beta_vec[2] ** 2.0,
                gamma * beta_vec[2],
            ],
            [gamma * beta_vec[0], gamma * beta_vec[1], gamma * beta_vec[2], gamma],
        ]
    )


def apply_transform(transform: np.ndarray, momentum: np.ndarray) -> np.ndarray:
    """Apply a Lorentz transform to a 4-vector

    Parameters
    ----------
    transform: ndarray
        A 4x4 matrix representing the transformation
    momentum: ndarray
        A length 4 array representing the momentum 4-vector

    Returns
    -------
    ndarray
        A new length 4 array representing the transformed 4-vector
    """
    return transform @ momentum.T
